@@ -3,10 +3,11 @@
 /* Some constants. */
 #define HEADER_SIZE                 92
 #define MIN_BYTECODE_VERSION        5
-#define MAX_BYTECODE_VERSION        5
-#define FRAME_HEADER_SIZE           (11 * 4 + 3 * 2)
+#define MAX_BYTECODE_VERSION        6
+#define FRAME_HEADER_SIZE           (11 * 4 + 3 * 2 + (bytecode_version >= 6 ? 4 : 0))
 #define FRAME_HANDLER_SIZE          (4 * 4 + 2 * 2)
 #define FRAME_SLV_SIZE              (2 * 2 + 2 * 4)
+#define FRAME_DEBUG_NAME_SIZE       (2 + 4)
 #define SCDEP_HEADER_OFFSET         12
 #define EXTOP_HEADER_OFFSET         20
 #define FRAME_HEADER_OFFSET         28
@@ -448,6 +449,7 @@ static MVMStaticFrame ** deserialize_frames(MVMThreadContext *tc, MVMCompUnit *c
     MVMStaticFrame **frames;
     MVMuint8        *pos;
     MVMuint32        bytecode_pos, bytecode_size, i, j;
+    MVMuint16        bytecode_version = rs->version;
 
     /* Allocate frames array. */
     if (rs->expected_frames == 0) {
@@ -476,7 +478,7 @@ static MVMStaticFrame ** deserialize_frames(MVMThreadContext *tc, MVMCompUnit *c
         bytecode_size = read_int32(pos, 4);
         if (bytecode_pos >= rs->bytecode_size) {
             cleanup_all(tc, rs);
-            MVM_exception_throw_adhoc(tc, "Frame has invalid bytecode start point");
+            MVM_exception_throw_adhoc(tc, "Frame has invalid bytecode start point %d (size %d)", bytecode_pos, rs->bytecode_size);
         }
         if (bytecode_pos + bytecode_size > rs->bytecode_size) {
             cleanup_all(tc, rs);
@@ -537,6 +539,7 @@ static MVMStaticFrame ** deserialize_frames(MVMThreadContext *tc, MVMCompUnit *c
             MVMuint32 skip = 2 * static_frame_body->num_locals +
                              6 * static_frame_body->num_lexicals;
             MVMuint16 slvs = read_int16(pos, 40);
+            MVMuint32 num_local_debug_names = rs->version >= 6 ? read_int32(pos, 50) : 0;
             pos += FRAME_HEADER_SIZE;
             ensure_can_read(tc, cu, rs, pos, skip);
             pos += skip;
@@ -553,6 +556,8 @@ static MVMStaticFrame ** deserialize_frames(MVMThreadContext *tc, MVMCompUnit *c
             }
             ensure_can_read(tc, cu, rs, pos, slvs * FRAME_SLV_SIZE);
             pos += slvs * FRAME_SLV_SIZE;
+            ensure_can_read(tc, cu, rs, pos, num_local_debug_names * FRAME_DEBUG_NAME_SIZE);
+            pos += num_local_debug_names * FRAME_DEBUG_NAME_SIZE;
         }
     }
 
@@ -575,9 +580,10 @@ static MVMStaticFrame ** deserialize_frames(MVMThreadContext *tc, MVMCompUnit *c
 /* Finishes up reading and exploding of a frame. */
 void MVM_bytecode_finish_frame(MVMThreadContext *tc, MVMCompUnit *cu,
                                MVMStaticFrame *sf, MVMint32 dump_only) {
-    MVMuint32 j;
+    MVMuint32 j, num_debug_locals;
     MVMuint8 *pos;
     MVMuint16 slvs;
+    MVMuint16 bytecode_version = cu->body.bytecode_version;
 
     /* Ensure we've not already done this. */
     if (sf->body.fully_deserialized)
@@ -599,8 +605,10 @@ void MVM_bytecode_finish_frame(MVMThreadContext *tc, MVMCompUnit *cu,
     /* Locate start of frame body. */
     pos = sf->body.frame_data_pos;
 
-    /* Get the number of static lex values we'll need to apply. */
+    /* Get the number of static lex values and debug local names we'll need
+     * to apply. */
     slvs = read_int16(pos, 40);
+    num_debug_locals = bytecode_version >= 6 ? read_int16(pos, 50) : 0;
 
     /* Skip past header. */
     pos += FRAME_HEADER_SIZE;
@@ -674,7 +682,7 @@ void MVM_bytecode_finish_frame(MVMThreadContext *tc, MVMCompUnit *cu,
 
         if (lex_idx >= sf->body.num_lexicals) {
             MVM_reentrantmutex_unlock(tc, (MVMReentrantMutex *)cu->body.deserialize_frame_mutex);
-            MVM_exception_throw_adhoc(tc, "Lexical index out of bounds: %d > %d", lex_idx, sf->body.num_lexicals);
+            MVM_exception_throw_adhoc(tc, "Lexical index out of bounds: %d >= %d", lex_idx, sf->body.num_lexicals);
         }
 
         sf->body.static_env_flags[lex_idx] = flags;
@@ -692,6 +700,23 @@ void MVM_bytecode_finish_frame(MVMThreadContext *tc, MVMCompUnit *cu,
         pos += FRAME_SLV_SIZE;
     }
 
+    /* Read in the debug local names if we're running the debug server. */
+    if (num_debug_locals && tc->instance->debugserver) {
+        MVMStaticFrameInstrumentation *ins = sf->body.instrumentation;
+        if (!ins)
+            ins = MVM_calloc(1, sizeof(MVMStaticFrameInstrumentation));
+        for (j = 0; j < num_debug_locals; j++) {
+            MVMuint16 idx = read_int16(pos, 0);
+            MVMString *name = get_heap_string(tc, cu, NULL, pos, 2);
+            MVMStaticFrameDebugLocal *entry = MVM_calloc(1, sizeof(MVMStaticFrameDebugLocal));
+            entry->local_idx = idx;
+            MVM_ASSIGN_REF(tc, &(sf->common.header), entry->name, name);
+            MVM_HASH_BIND(tc, ins->debug_locals, name, entry);
+            pos += 6;
+        }
+        sf->body.instrumentation = ins;
+    }
+
     /* Mark the frame fully deserialized. */
     sf->body.fully_deserialized = 1;
 
@@ -701,7 +726,7 @@ void MVM_bytecode_finish_frame(MVMThreadContext *tc, MVMCompUnit *cu,
 
 /* Gets the SC reference for a given static lexical var for
  * vivification purposes */
-MVMuint8 MVM_bytecode_find_static_lexical_scref(MVMThreadContext *tc, MVMCompUnit *cu, MVMStaticFrame *sf, MVMuint16 index, MVMint32 *sc, MVMint32 *id) {
+MVMuint8 MVM_bytecode_find_static_lexical_scref(MVMThreadContext *tc, MVMCompUnit *cu, MVMStaticFrame *sf, MVMuint16 index, MVMuint32 *sc, MVMuint32 *id) {
     MVMuint16 slvs, i;
 
     MVMuint8 *pos = sf->body.frame_static_lex_pos;
@@ -786,7 +811,7 @@ static MVMCallsite ** deserialize_callsites(MVMThreadContext *tc, MVMCompUnit *c
                 nameds_non_flattening++;
             }
             else if (nameds_slots) {
-                MVM_exception_throw_adhoc(tc, "All positional args must appear before named args");
+                MVM_exception_throw_adhoc(tc, "All positional args must appear before named args, violated by arg %d in callsite %d", j, i);
             }
             else {
                 positionals++;
@@ -918,7 +943,7 @@ MVMBytecodeAnnotation * MVM_bytecode_resolve_annotation(MVMThreadContext *tc, MV
     if (sfb->num_annotations && offset < sfb->bytecode_size) {
         MVMuint8 *cur_anno = sfb->annotations_data;
         for (i = 0; i < sfb->num_annotations; i++) {
-            MVMint32 ann_offset = read_int32(cur_anno, 0);
+            MVMuint32 ann_offset = read_int32(cur_anno, 0);
             if (ann_offset > offset)
                 break;
             cur_anno += 12;
