@@ -54,6 +54,7 @@ static char get_signature_char(MVMint16 type_id) {
         case MVM_NATIVECALL_ARG_UTF8STR:
         case MVM_NATIVECALL_ARG_UTF16STR:
         case MVM_NATIVECALL_ARG_CSTRUCT:
+        case MVM_NATIVECALL_ARG_CPPSTRUCT:
         case MVM_NATIVECALL_ARG_CPOINTER:
         case MVM_NATIVECALL_ARG_CARRAY:
         case MVM_NATIVECALL_ARG_CUNION:
@@ -179,10 +180,12 @@ static void * unmarshal_callback(MVMThreadContext *tc, MVMObject *callback, MVMO
 
         MVM_callsite_try_intern(tc, &cs);
 
-        callback_data->tc        = tc;
+        callback_data->instance  = tc->instance;
         callback_data->cs        = cs;
         callback_data->target    = callback;
         callback_data->cb        = dcbNewCallback(signature, (DCCallbackHandler *)callback_handler, callback_data);
+        if (!callback_data->cb)
+            MVM_panic(1, "Unable to allocate memory for callback closure");
 
         /* Now insert the MVMCallback into the linked list. */
         *callback_data_handle = callback_data;
@@ -209,15 +212,20 @@ static void callback_invoke(MVMThreadContext *tc, void *data) {
 static char callback_handler(DCCallback *cb, DCArgs *cb_args, DCValue *cb_result, MVMNativeCallback *data) {
     CallbackInvokeData cid;
     MVMint32 num_roots, i;
-    MVMRegister res;
+    MVMRegister res = {0};
     MVMRegister *args;
+    unsigned int interval_id;
+
+    /* Locate the MoarVM thread this callback is being run on. */
+    MVMThreadContext *tc = MVM_nativecall_find_thread_context(data->instance);
 
     /* Unblock GC if needed, so this thread can do work. */
-    MVMThreadContext *tc = data->tc;
     MVMint32 was_blocked = MVM_gc_is_thread_blocked(tc);
     if (was_blocked)
         MVM_gc_mark_thread_unblocked(tc);
 
+    interval_id = MVM_telemetry_interval_start(tc, "nativecall callback handler");
+
     /* Build a callsite and arguments buffer. */
     args = MVM_malloc(data->num_types * sizeof(MVMRegister));
     num_roots = 0;
@@ -260,6 +268,12 @@ static char callback_handler(DCCallback *cb, DCArgs *cb_args, DCValue *cb_result
                 MVM_gc_root_temp_push(tc, (MVMCollectable **)&(args[i - 1].o));
                 num_roots++;
                 break;
+            case MVM_NATIVECALL_ARG_CPPSTRUCT:
+                args[i - 1].o = MVM_nativecall_make_cppstruct(tc, type,
+                    dcbArgPointer(cb_args));
+                MVM_gc_root_temp_push(tc, (MVMCollectable **)&(args[i - 1].o));
+                num_roots++;
+                break;
             case MVM_NATIVECALL_ARG_CPOINTER:
                 args[i - 1].o = MVM_nativecall_make_cpointer(tc, type,
                     dcbArgPointer(cb_args));
@@ -303,6 +317,7 @@ static char callback_handler(DCCallback *cb, DCArgs *cb_args, DCValue *cb_result
                 args[i - 1].i64 = dcbArgULongLong(cb_args);
                 break;
             default:
+                MVM_telemetry_interval_stop(tc, interval_id, "nativecall callback handler failed");
                 MVM_exception_throw_adhoc(tc,
                     "Internal error: unhandled dyncall callback argument type");
         }
@@ -313,106 +328,82 @@ static char callback_handler(DCCallback *cb, DCArgs *cb_args, DCValue *cb_result
     cid.invokee = data->target;
     cid.args    = args;
     cid.cs      = data->cs;
-    {
-        MVMuint8 **backup_interp_cur_op         = tc->interp_cur_op;
-        MVMuint8 **backup_interp_bytecode_start = tc->interp_bytecode_start;
-        MVMRegister **backup_interp_reg_base    = tc->interp_reg_base;
-        MVMCompUnit **backup_interp_cu          = tc->interp_cu;
-        MVMFrame *backup_cur_frame              = MVM_frame_force_to_heap(tc, tc->cur_frame);
-        MVMFrame *backup_thread_entry_frame     = tc->thread_entry_frame;
-        MVMROOT(tc, backup_cur_frame, {
-        MVMROOT(tc, backup_thread_entry_frame, {
-            MVMuint32 backup_mark                   = MVM_gc_root_temp_mark(tc);
-            jmp_buf backup_interp_jump;
-            memcpy(backup_interp_jump, tc->interp_jump, sizeof(jmp_buf));
-
-            tc->cur_frame->return_value = &res;
-            tc->cur_frame->return_type  = MVM_RETURN_OBJ;
-            MVM_interp_run(tc, callback_invoke, &cid);
-
-            tc->interp_cur_op         = backup_interp_cur_op;
-            tc->interp_bytecode_start = backup_interp_bytecode_start;
-            tc->interp_reg_base       = backup_interp_reg_base;
-            tc->interp_cu             = backup_interp_cu;
-            tc->cur_frame             = backup_cur_frame;
-            tc->current_frame_nr      = backup_cur_frame->sequence_nr;
-            tc->thread_entry_frame    = backup_thread_entry_frame;
-            memcpy(tc->interp_jump, backup_interp_jump, sizeof(jmp_buf));
-            MVM_gc_root_temp_mark_reset(tc, backup_mark);
-        });
-        });
-    }
+    MVM_interp_run_nested(tc, callback_invoke, &cid, &res);
 
     /* Handle return value. */
     if (res.o) {
         MVMContainerSpec const *contspec = STABLE(res.o)->container_spec;
         if (contspec && contspec->fetch_never_invokes)
-            contspec->fetch(data->tc, res.o, &res);
+            contspec->fetch(tc, res.o, &res);
     }
     switch (data->typeinfos[0] & MVM_NATIVECALL_ARG_TYPE_MASK) {
         case MVM_NATIVECALL_ARG_VOID:
             break;
         case MVM_NATIVECALL_ARG_CHAR:
-            cb_result->c = (signed char)MVM_nativecall_unmarshal_char(data->tc, res.o);
+            cb_result->c = (signed char)MVM_nativecall_unmarshal_char(tc, res.o);
             break;
         case MVM_NATIVECALL_ARG_SHORT:
-            cb_result->s = MVM_nativecall_unmarshal_short(data->tc, res.o);
+            cb_result->s = MVM_nativecall_unmarshal_short(tc, res.o);
             break;
         case MVM_NATIVECALL_ARG_INT:
-            cb_result->i = MVM_nativecall_unmarshal_int(data->tc, res.o);
+            cb_result->i = MVM_nativecall_unmarshal_int(tc, res.o);
             break;
         case MVM_NATIVECALL_ARG_LONG:
-            cb_result->j = MVM_nativecall_unmarshal_long(data->tc, res.o);
+            cb_result->j = MVM_nativecall_unmarshal_long(tc, res.o);
             break;
         case MVM_NATIVECALL_ARG_LONGLONG:
-            cb_result->l = MVM_nativecall_unmarshal_longlong(data->tc, res.o);
+            cb_result->l = MVM_nativecall_unmarshal_longlong(tc, res.o);
             break;
         case MVM_NATIVECALL_ARG_FLOAT:
-            cb_result->f = MVM_nativecall_unmarshal_float(data->tc, res.o);
+            cb_result->f = MVM_nativecall_unmarshal_float(tc, res.o);
             break;
         case MVM_NATIVECALL_ARG_DOUBLE:
-            cb_result->d = MVM_nativecall_unmarshal_double(data->tc, res.o);
+            cb_result->d = MVM_nativecall_unmarshal_double(tc, res.o);
             break;
         case MVM_NATIVECALL_ARG_ASCIISTR:
         case MVM_NATIVECALL_ARG_UTF8STR:
         case MVM_NATIVECALL_ARG_UTF16STR:
-            cb_result->Z = MVM_nativecall_unmarshal_string(data->tc, res.o, data->typeinfos[0], NULL);
+            cb_result->Z = MVM_nativecall_unmarshal_string(tc, res.o, data->typeinfos[0], NULL, MVM_NATIVECALL_UNMARSHAL_KIND_GENERIC);
             break;
         case MVM_NATIVECALL_ARG_CSTRUCT:
-            cb_result->p = MVM_nativecall_unmarshal_cstruct(data->tc, res.o);
+            cb_result->p = MVM_nativecall_unmarshal_cstruct(tc, res.o, MVM_NATIVECALL_UNMARSHAL_KIND_GENERIC);
+            break;
+        case MVM_NATIVECALL_ARG_CPPSTRUCT:
+            cb_result->p = MVM_nativecall_unmarshal_cppstruct(tc, res.o, MVM_NATIVECALL_UNMARSHAL_KIND_GENERIC);
             break;
         case MVM_NATIVECALL_ARG_CPOINTER:
-            cb_result->p = MVM_nativecall_unmarshal_cpointer(data->tc, res.o);
+            cb_result->p = MVM_nativecall_unmarshal_cpointer(tc, res.o, MVM_NATIVECALL_UNMARSHAL_KIND_GENERIC);
             break;
         case MVM_NATIVECALL_ARG_CARRAY:
-            cb_result->p = MVM_nativecall_unmarshal_carray(data->tc, res.o);
+            cb_result->p = MVM_nativecall_unmarshal_carray(tc, res.o, MVM_NATIVECALL_UNMARSHAL_KIND_GENERIC);
             break;
         case MVM_NATIVECALL_ARG_CUNION:
-            cb_result->p = MVM_nativecall_unmarshal_cunion(data->tc, res.o);
+            cb_result->p = MVM_nativecall_unmarshal_cunion(tc, res.o, MVM_NATIVECALL_UNMARSHAL_KIND_GENERIC);
             break;
         case MVM_NATIVECALL_ARG_VMARRAY:
-            cb_result->p = MVM_nativecall_unmarshal_vmarray(data->tc, res.o);
+            cb_result->p = MVM_nativecall_unmarshal_vmarray(tc, res.o, MVM_NATIVECALL_UNMARSHAL_KIND_GENERIC);
             break;
         case MVM_NATIVECALL_ARG_CALLBACK:
-            cb_result->p = unmarshal_callback(data->tc, res.o, data->types[0]);
+            cb_result->p = unmarshal_callback(tc, res.o, data->types[0]);
             break;
         case MVM_NATIVECALL_ARG_UCHAR:
-            cb_result->c = MVM_nativecall_unmarshal_uchar(data->tc, res.o);
+            cb_result->c = MVM_nativecall_unmarshal_uchar(tc, res.o);
             break;
         case MVM_NATIVECALL_ARG_USHORT:
-            cb_result->s = MVM_nativecall_unmarshal_ushort(data->tc, res.o);
+            cb_result->s = MVM_nativecall_unmarshal_ushort(tc, res.o);
             break;
         case MVM_NATIVECALL_ARG_UINT:
-            cb_result->i = MVM_nativecall_unmarshal_uint(data->tc, res.o);
+            cb_result->i = MVM_nativecall_unmarshal_uint(tc, res.o);
             break;
         case MVM_NATIVECALL_ARG_ULONG:
-            cb_result->j = MVM_nativecall_unmarshal_ulong(data->tc, res.o);
+            cb_result->j = MVM_nativecall_unmarshal_ulong(tc, res.o);
             break;
         case MVM_NATIVECALL_ARG_ULONGLONG:
-            cb_result->l = MVM_nativecall_unmarshal_ulonglong(data->tc, res.o);
+            cb_result->l = MVM_nativecall_unmarshal_ulonglong(tc, res.o);
             break;
         default:
-            MVM_exception_throw_adhoc(data->tc,
+            MVM_telemetry_interval_stop(tc, interval_id, "nativecall callback handler failed");
+            MVM_exception_throw_adhoc(tc,
                 "Internal error: unhandled dyncall callback return type");
     }
 
@@ -424,6 +415,8 @@ static char callback_handler(DCCallback *cb, DCArgs *cb_args, DCValue *cb_result
     if (was_blocked)
         MVM_gc_mark_thread_blocked(tc);
 
+    MVM_telemetry_interval_stop(tc, interval_id, "nativecall callback handler");
+
     /* Indicate what we're producing as a result. */
     return get_signature_char(data->typeinfos[0]);
 }
@@ -443,8 +436,8 @@ static char callback_handler(DCCallback *cb, DCArgs *cb_args, DCValue *cb_result
         } \
         else \
             MVM_exception_throw_adhoc(tc, \
-                "Native call expected argument that references a native %s, but got %s", \
-                what, REPR(value)->name); \
+                "Native call expected argument %d to reference a native %s, but got %s", \
+                i, what, REPR(value)->name); \
     } \
     else { \
         if (value && IS_CONCRETE(value) && STABLE(value)->container_spec) { \
@@ -469,17 +462,29 @@ MVMObject * MVM_nativecall_invoke(MVMThreadContext *tc, MVMObject *res_type,
     /* Get native call body, so we can locate the call info. Read out all we
      * shall need, since later we may allocate a result and and move it. */
     MVMNativeCallBody *body = MVM_nativecall_get_nc_body(tc, site);
+    if (MVM_UNLIKELY(!body->lib_handle)) {
+        MVMROOT3(tc, site, args, res_type, {
+            MVM_nativecall_restore_library(tc, body, site);
+        });
+        body = MVM_nativecall_get_nc_body(tc, site);
+    }
     MVMint16  num_args    = body->num_args;
     MVMint16 *arg_types   = body->arg_types;
     MVMint16  ret_type    = body->ret_type;
     void     *entry_point = body->entry_point;
     void     *ptr         = NULL;
 
+    unsigned int interval_id;
+    DCCallVM *vm;
+
     /* Create and set up call VM. */
-    DCCallVM *vm = dcNewCallVM(8192);
+    vm = dcNewCallVM(8192);
     dcMode(vm, body->convention);
     dcReset(vm);
 
+    interval_id = MVM_telemetry_interval_start(tc, "nativecall invoke");
+    MVM_telemetry_interval_annotate((intptr_t)entry_point, interval_id, "nc entrypoint");
+
     /* Process arguments. */
     for (i = 0; i < num_args; i++) {
         MVMObject *value = MVM_repr_at_pos_o(tc, args, i);
@@ -510,7 +515,7 @@ MVMObject * MVM_nativecall_invoke(MVMThreadContext *tc, MVMObject *res_type,
             case MVM_NATIVECALL_ARG_UTF16STR:
                 {
                     MVMint16 free = 0;
-                    char *str = MVM_nativecall_unmarshal_string(tc, value, arg_types[i], &free);
+                    char *str = MVM_nativecall_unmarshal_string(tc, value, arg_types[i], &free, i);
                     if (free) {
                         if (!free_strs)
                             free_strs = (char**)MVM_malloc(num_args * sizeof(char *));
@@ -521,7 +526,7 @@ MVMObject * MVM_nativecall_invoke(MVMThreadContext *tc, MVMObject *res_type,
                 }
                 break;
             case MVM_NATIVECALL_ARG_CSTRUCT:
-                dcArgPointer(vm, MVM_nativecall_unmarshal_cstruct(tc, value));
+                dcArgPointer(vm, MVM_nativecall_unmarshal_cstruct(tc, value, i));
                 break;
             case MVM_NATIVECALL_ARG_CPPSTRUCT: {
                     /* We need to allocate the struct (THIS) for C++ constructor before passing it along. */
@@ -534,14 +539,14 @@ MVMObject * MVM_nativecall_invoke(MVMThreadContext *tc, MVMObject *res_type,
                         dcArgPointer(vm, ptr);
                     }
                     else {
-                        dcArgPointer(vm, MVM_nativecall_unmarshal_cppstruct(tc, value));
+                        dcArgPointer(vm, MVM_nativecall_unmarshal_cppstruct(tc, value, i));
                     }
                 }
                 break;
             case MVM_NATIVECALL_ARG_CPOINTER:
                 if ((arg_types[i] & MVM_NATIVECALL_ARG_RW_MASK) == MVM_NATIVECALL_ARG_RW) {
                     DCpointer *rw = (DCpointer *)MVM_malloc(sizeof(DCpointer *));
-                    *rw           = (DCpointer)MVM_nativecall_unmarshal_cpointer(tc, value);
+                    *rw           = (DCpointer)MVM_nativecall_unmarshal_cpointer(tc, value, i);
                     if (!free_rws)
                         free_rws = (void **)MVM_malloc(num_args * sizeof(void *));
                     free_rws[num_rws] = rw;
@@ -549,17 +554,17 @@ MVMObject * MVM_nativecall_invoke(MVMThreadContext *tc, MVMObject *res_type,
                     dcArgPointer(vm, rw);
                 }
                 else {
-                    dcArgPointer(vm, MVM_nativecall_unmarshal_cpointer(tc, value));
+                    dcArgPointer(vm, MVM_nativecall_unmarshal_cpointer(tc, value, i));
                 }
                 break;
             case MVM_NATIVECALL_ARG_CARRAY:
-                dcArgPointer(vm, MVM_nativecall_unmarshal_carray(tc, value));
+                dcArgPointer(vm, MVM_nativecall_unmarshal_carray(tc, value, i));
                 break;
             case MVM_NATIVECALL_ARG_CUNION:
-                dcArgPointer(vm, MVM_nativecall_unmarshal_cunion(tc, value));
+                dcArgPointer(vm, MVM_nativecall_unmarshal_cunion(tc, value, i));
                 break;
             case MVM_NATIVECALL_ARG_VMARRAY:
-                dcArgPointer(vm, MVM_nativecall_unmarshal_vmarray(tc, value));
+                dcArgPointer(vm, MVM_nativecall_unmarshal_vmarray(tc, value, i));
                 break;
             case MVM_NATIVECALL_ARG_CALLBACK:
                 dcArgPointer(vm, unmarshal_callback(tc, value, body->arg_info[i]));
@@ -580,12 +585,12 @@ MVMObject * MVM_nativecall_invoke(MVMThreadContext *tc, MVMObject *res_type,
                 handle_arg("integer", cont_i, DCulonglong, i64, dcArgLongLong, MVM_nativecall_unmarshal_ulonglong);
                 break;
             default:
+                MVM_telemetry_interval_stop(tc, interval_id, "nativecall invoke failed");
                 MVM_exception_throw_adhoc(tc, "Internal error: unhandled dyncall argument type");
         }
     }
 
-    MVMROOT(tc, args, {
-    MVMROOT(tc, res_type, {
+    MVMROOT2(tc, args, res_type, {
         MVM_gc_mark_thread_blocked(tc);
         if (result) {
             /* We are calling a C++ constructor so we hand back the invocant (THIS) we recorded earlier. */
@@ -720,11 +725,11 @@ MVMObject * MVM_nativecall_invoke(MVMThreadContext *tc, MVMObject *res_type,
                     break;
                 }
                 default:
+                    MVM_telemetry_interval_stop(tc, interval_id, "nativecall invoke failed");
                     MVM_exception_throw_adhoc(tc, "Internal error: unhandled dyncall return type");
             }
         }
     });
-    });
 
     num_rws = 0;
     for (i = 0; i < num_args; i++) {
@@ -772,6 +777,7 @@ MVMObject * MVM_nativecall_invoke(MVMThreadContext *tc, MVMObject *res_type,
                         (MVMint64)*(DCpointer *)free_rws[num_rws]);
                     break;
                 default:
+                    MVM_telemetry_interval_stop(tc, interval_id, "nativecall invoke failed");
                     MVM_exception_throw_adhoc(tc, "Internal error: unhandled dyncall argument type");
             }
             num_rws++;
@@ -796,5 +802,23 @@ MVMObject * MVM_nativecall_invoke(MVMThreadContext *tc, MVMObject *res_type,
     /* Finally, free call VM. */
     dcFree(vm);
 
+    MVM_telemetry_interval_stop(tc, interval_id, "nativecall invoke");
     return result;
 }
+
+#ifdef _WIN32
+DLLib* MVM_nativecall_load_lib(const char* path) {
+    HMODULE module;
+    if (path != NULL) {
+        const int       len   = MultiByteToWideChar(CP_UTF8, 0, path, -1, NULL, 0);
+        wchar_t * const wpath = (wchar_t *)MVM_malloc(len * sizeof(wchar_t));
+                                MultiByteToWideChar(CP_UTF8, 0, path, -1, (LPWSTR)wpath, len);
+        module = LoadLibraryW(wpath);
+        MVM_free(wpath);
+    }
+    else {
+        module = GetModuleHandle(NULL);
+    }
+    return (DLLib*)(module);
+}
+#endif
