@@ -28,7 +28,8 @@ MVM_STATIC_INLINE void adjust_nursery(MVMThreadContext *tc, MVMP6bigintBody *bod
 
 /* Taken from mp_set_long, but portably accepts a 64-bit number. */
 int MVM_bigint_mp_set_uint64(mp_int * a, MVMuint64 b) {
-  int     x, res;
+  size_t x;
+  int    res;
 
   mp_zero (a);
 
@@ -426,31 +427,32 @@ MVMObject * MVM_bigint_##opname(MVMThreadContext *tc, MVMObject *result_type, MV
 
 #define MVM_BIGINT_BINARY_OP_2(opname, SMALLINT_OP) \
 MVMObject * MVM_bigint_##opname(MVMThreadContext *tc, MVMObject *result_type, MVMObject *a, MVMObject *b) { \
-    MVMP6bigintBody *ba = get_bigint_body(tc, a); \
-    MVMP6bigintBody *bb = get_bigint_body(tc, b); \
-    MVMP6bigintBody *bc; \
     MVMObject *result; \
     MVMROOT2(tc, a, b, { \
         result = MVM_repr_alloc_init(tc, result_type);\
     }); \
-    bc = get_bigint_body(tc, result); \
-    if (MVM_BIGINT_IS_BIG(ba) || MVM_BIGINT_IS_BIG(bb)) { \
-        mp_int *ia = force_bigint(tc, ba, 0); \
-        mp_int *ib = force_bigint(tc, bb, 1); \
-        mp_int *ic = MVM_malloc(sizeof(mp_int)); \
-        mp_init(ic); \
-        two_complement_bitop(ia, ib, ic, mp_##opname); \
-        store_bigint_result(bc, ic); \
-        adjust_nursery(tc, bc); \
-    } \
-    else { \
-        MVMint64 sc; \
-        MVMint64 sa = ba->u.smallint.value; \
-        MVMint64 sb = bb->u.smallint.value; \
-        SMALLINT_OP; \
-        store_int64_result(bc, sc); \
-    } \
-    return result; \
+    {\
+        MVMP6bigintBody *ba = get_bigint_body(tc, a); \
+        MVMP6bigintBody *bb = get_bigint_body(tc, b); \
+        MVMP6bigintBody *bc = get_bigint_body(tc, result); \
+        if (MVM_BIGINT_IS_BIG(ba) || MVM_BIGINT_IS_BIG(bb)) { \
+            mp_int *ia = force_bigint(tc, ba, 0); \
+            mp_int *ib = force_bigint(tc, bb, 1); \
+            mp_int *ic = MVM_malloc(sizeof(mp_int)); \
+            mp_init(ic); \
+            two_complement_bitop(ia, ib, ic, mp_##opname); \
+            store_bigint_result(bc, ic); \
+            adjust_nursery(tc, bc); \
+        } \
+        else { \
+            MVMint64 sc; \
+            MVMint64 sa = ba->u.smallint.value; \
+            MVMint64 sb = bb->u.smallint.value; \
+            SMALLINT_OP; \
+            store_int64_result(bc, sc); \
+        } \
+        return result; \
+    }\
 }
 
 MVM_BIGINT_UNARY_OP(abs, { sb = labs(sa); })
@@ -465,37 +467,38 @@ MVM_BIGINT_BINARY_OP_SIMPLE(mul, { sc = sa * sb; })
 MVM_BIGINT_BINARY_OP(lcm)
 
 MVMObject *MVM_bigint_gcd(MVMThreadContext *tc, MVMObject *result_type, MVMObject *a, MVMObject *b) {
-    MVMP6bigintBody *ba = get_bigint_body(tc, a);
-    MVMP6bigintBody *bb = get_bigint_body(tc, b);
-    MVMP6bigintBody *bc;
     MVMObject       *result;
 
     MVMROOT2(tc, a, b, {
         result = MVM_repr_alloc_init(tc, result_type);
     });
 
-    bc = get_bigint_body(tc, result);
+    {
+        MVMP6bigintBody *ba = get_bigint_body(tc, a);
+        MVMP6bigintBody *bb = get_bigint_body(tc, b);
+        MVMP6bigintBody *bc = get_bigint_body(tc, result);
 
-    if (MVM_BIGINT_IS_BIG(ba) || MVM_BIGINT_IS_BIG(bb)) {
-        mp_int *ia = force_bigint(tc, ba, 0);
-        mp_int *ib = force_bigint(tc, bb, 1);
-        mp_int *ic = MVM_malloc(sizeof(mp_int));
-        mp_init(ic);
-        mp_gcd(ia, ib, ic);
-        store_bigint_result(bc, ic);
-        adjust_nursery(tc, bc);
-    } else {
-        MVMint32 sa = ba->u.smallint.value;
-        MVMint32 sb = bb->u.smallint.value;
-        MVMint32 t;
-        sa = abs(sa);
-        sb = abs(sb);
-        while (sb != 0) {
-            t  = sb;
-            sb = sa % sb;
-            sa = t;
+        if (MVM_BIGINT_IS_BIG(ba) || MVM_BIGINT_IS_BIG(bb)) {
+            mp_int *ia = force_bigint(tc, ba, 0);
+            mp_int *ib = force_bigint(tc, bb, 1);
+            mp_int *ic = MVM_malloc(sizeof(mp_int));
+            mp_init(ic);
+            mp_gcd(ia, ib, ic);
+            store_bigint_result(bc, ic);
+            adjust_nursery(tc, bc);
+        } else {
+            MVMint32 sa = ba->u.smallint.value;
+            MVMint32 sb = bb->u.smallint.value;
+            MVMint32 t;
+            sa = abs(sa);
+            sb = abs(sb);
+            while (sb != 0) {
+                t  = sb;
+                sb = sa % sb;
+                sa = t;
+            }
+            store_int64_result(bc, sa);
         }
-        store_int64_result(bc, sa);
     }
 
     return result;
@@ -522,9 +525,6 @@ MVMint64 MVM_bigint_cmp(MVMThreadContext *tc, MVMObject *a, MVMObject *b) {
 }
 
 MVMObject * MVM_bigint_mod(MVMThreadContext *tc, MVMObject *result_type, MVMObject *a, MVMObject *b) {
-    MVMP6bigintBody *ba = get_bigint_body(tc, a);
-    MVMP6bigintBody *bb = get_bigint_body(tc, b);
-    MVMP6bigintBody *bc;
 
     MVMObject *result;
 
@@ -532,35 +532,40 @@ MVMObject * MVM_bigint_mod(MVMThreadContext *tc, MVMObject *result_type, MVMObje
         result = MVM_repr_alloc_init(tc, result_type);
     });
 
-    bc = get_bigint_body(tc, result);
+    {
+        MVMP6bigintBody *ba = get_bigint_body(tc, a);
+        MVMP6bigintBody *bb = get_bigint_body(tc, b);
+        MVMP6bigintBody *bc;
+        bc = get_bigint_body(tc, result);
 
-    /* XXX the behavior of C's mod operator is not correct
-     * for our purposes. So we rely on mp_mod for all our modulus
-     * calculations for now. */
-    if (1 || MVM_BIGINT_IS_BIG(ba) || MVM_BIGINT_IS_BIG(bb)) {
-        mp_int *ia = force_bigint(tc, ba, 0);
-        mp_int *ib = force_bigint(tc, bb, 1);
-        mp_int *ic = MVM_malloc(sizeof(mp_int));
-        int mp_result;
+        /* XXX the behavior of C's mod operator is not correct
+         * for our purposes. So we rely on mp_mod for all our modulus
+         * calculations for now. */
+        if (1 || MVM_BIGINT_IS_BIG(ba) || MVM_BIGINT_IS_BIG(bb)) {
+            mp_int *ia = force_bigint(tc, ba, 0);
+            mp_int *ib = force_bigint(tc, bb, 1);
+            mp_int *ic = MVM_malloc(sizeof(mp_int));
+            int mp_result;
 
-        mp_init(ic);
+            mp_init(ic);
 
-        mp_result = mp_mod(ia, ib, ic);
+            mp_result = mp_mod(ia, ib, ic);
 
-        if (mp_result == MP_VAL) {
-            MVM_exception_throw_adhoc(tc, "Division by zero");
+            if (mp_result == MP_VAL) {
+                MVM_exception_throw_adhoc(tc, "Division by zero");
+            }
+            store_bigint_result(bc, ic);
+            adjust_nursery(tc, bc);
+        } else {
+            store_int64_result(bc, ba->u.smallint.value % bb->u.smallint.value);
         }
-        store_bigint_result(bc, ic);
-        adjust_nursery(tc, bc);
-    } else {
-        store_int64_result(bc, ba->u.smallint.value % bb->u.smallint.value);
     }
 
     return result;
 }
 
 MVMObject *MVM_bigint_div(MVMThreadContext *tc, MVMObject *result_type, MVMObject *a, MVMObject *b) {
-    MVMP6bigintBody *ba = get_bigint_body(tc, a);
+    MVMP6bigintBody *ba;
     MVMP6bigintBody *bb = get_bigint_body(tc, b);
     MVMP6bigintBody *bc;
     mp_int *ia, *ib, *ic;
@@ -580,6 +585,9 @@ MVMObject *MVM_bigint_div(MVMThreadContext *tc, MVMObject *result_type, MVMObjec
         result = MVM_repr_alloc_init(tc, result_type);
     });
 
+    ba = get_bigint_body(tc, a);
+    bb = get_bigint_body(tc, b);
+
     bc = get_bigint_body(tc, result);
 
     /* we only care about MP_LT or !MP_LT, so we give MP_GT even for 0. */
@@ -705,7 +713,7 @@ MVMObject * MVM_bigint_pow(MVMThreadContext *tc, MVMObject *a, MVMObject *b,
 }
 
 MVMObject *MVM_bigint_shl(MVMThreadContext *tc, MVMObject *result_type, MVMObject *a, MVMint64 n) {
-    MVMP6bigintBody *ba = get_bigint_body(tc, a);
+    MVMP6bigintBody *ba;
     MVMP6bigintBody *bb;
     MVMObject       *result;
 
@@ -713,6 +721,7 @@ MVMObject *MVM_bigint_shl(MVMThreadContext *tc, MVMObject *result_type, MVMObjec
         result = MVM_repr_alloc_init(tc, result_type);
     });
 
+    ba = get_bigint_body(tc, a);
     bb = get_bigint_body(tc, result);
 
     if (MVM_BIGINT_IS_BIG(ba) || n >= 31) {
@@ -745,7 +754,7 @@ int BIGINT_IS_NEGATIVE (MVMP6bigintBody *ba) {
     }
 }
 MVMObject *MVM_bigint_shr(MVMThreadContext *tc, MVMObject *result_type, MVMObject *a, MVMint64 n) {
-    MVMP6bigintBody *ba = get_bigint_body(tc, a);
+    MVMP6bigintBody *ba;
     MVMP6bigintBody *bb;
     MVMObject       *result;
 
@@ -753,6 +762,7 @@ MVMObject *MVM_bigint_shr(MVMThreadContext *tc, MVMObject *result_type, MVMObjec
         result = MVM_repr_alloc_init(tc, result_type);
     });
 
+    ba = get_bigint_body(tc, a);
     bb = get_bigint_body(tc, result);
 
     if (MVM_BIGINT_IS_BIG(ba) || n < 0) {
@@ -774,7 +784,7 @@ MVMObject *MVM_bigint_shr(MVMThreadContext *tc, MVMObject *result_type, MVMObjec
 }
 
 MVMObject *MVM_bigint_not(MVMThreadContext *tc, MVMObject *result_type, MVMObject *a) {
-    MVMP6bigintBody *ba = get_bigint_body(tc, a);
+    MVMP6bigintBody *ba;
     MVMP6bigintBody *bb;
     MVMObject       *result;
 
@@ -782,6 +792,7 @@ MVMObject *MVM_bigint_not(MVMThreadContext *tc, MVMObject *result_type, MVMObjec
         result = MVM_repr_alloc_init(tc, result_type);
     });
 
+    ba = get_bigint_body(tc, a);
     bb = get_bigint_body(tc, result);
 
     if (MVM_BIGINT_IS_BIG(ba)) {
@@ -850,7 +861,10 @@ MVMObject * MVM_coerce_sI(MVMThreadContext *tc, MVMString *s, MVMObject *type) {
     char *buf = NULL;
     int is_malloced = 0;
     MVMStringIndex i;
-    MVMObject *a = MVM_repr_alloc_init(tc, type);
+    MVMObject *a;
+    MVMROOT(tc, s, {
+        a = MVM_repr_alloc_init(tc, type);
+    });
     if (s->body.num_graphs < 120) {
         buf = alloca(s->body.num_graphs + 1);
     }
@@ -1273,8 +1287,8 @@ MVMObject * MVM_bigint_radix(MVMThreadContext *tc, MVMint64 radix, MVMString *st
     mp_set_int(base, 1);
 
     ch = (offset < chars) ? MVM_string_get_grapheme_at_nocheck(tc, str, offset) : 0;
-    if ((flag & 0x02) && (ch == '+' || ch == '-')) {
-        neg = (ch == '-');
+    if ((flag & 0x02) && (ch == '+' || ch == '-' || ch == 0x2212)) {  /* MINUS SIGN */
+        neg = (ch == '-' || ch == 0x2212);
         offset++;
         ch = (offset < chars) ? MVM_string_get_grapheme_at_nocheck(tc, str, offset) : 0;
     }
